/* initng-conf-gtk, the gtk configuration program for initng
 * Copyright (C) 2006 Daniel Malmgren <dm@mensa.se>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#define _GNU_SOURCE

#include <gtk/gtk.h>
#include <glade/glade.h>

#include "serviceparser.h"
#include "initng-conf-gtk_globals.h"
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>
#include <fnmatch.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BASEDIR "/etc/initng"

int filefilter(const struct dirent* entry) {
  if(strcmp(entry->d_name, ".") && strcmp(entry->d_name, ".."))
    return 1;
  else
    return 0;
}

gboolean is_active(GtkTreeModel *model, GtkTreeIter *iter) {
  gchar *state;
  gtk_tree_model_get(model, iter, 2, &state, -1);
  if(state) {
    g_free(state);
    return TRUE;
  } else {
    g_free(state);
    return FALSE;
  }
}

gboolean is_failed(GtkTreeModel *model, GtkTreeIter *iter) {
  gchar *color;
  gtk_tree_model_get(model, iter, 3, &color, -1);
  if(color) {
    if(!strcmp(color, "#FFAAAA")) {
      g_free(color);
      return TRUE;
    } else {
      g_free(color);
    }
  }
  return FALSE;
}

gboolean recursive_filter(GtkTreeModel *model, GtkTreeIter *parent_iter,
			  gboolean (*testfunc)(GtkTreeModel*, GtkTreeIter*)) {
  GtkTreeIter iter;

  if(testfunc(model, parent_iter))
    return TRUE;

  gboolean valid = gtk_tree_model_iter_children(model, &iter, parent_iter);
  while(valid) {
    if(recursive_filter(model, &iter, testfunc)) {
      return TRUE;
    }
    valid = gtk_tree_model_iter_next(model, &iter);
  }
  return FALSE;
}

gboolean service_filter(GtkTreeModel *model, GtkTreeIter *iter, gpointer data) {
  GtkComboBox *filter_combo = GTK_COMBO_BOX(glade_xml_get_widget(xml,
								 "runlevel_filter_combo"));
  gint choosen_filter = gtk_combo_box_get_active(GTK_COMBO_BOX(filter_combo));
  gchar *str_data;

  //Just some example filters to test what can be done
  switch(choosen_filter) {
  case 1: //Show only active
    return recursive_filter(model, iter, is_active);
  case 2: //Show only failed
    return recursive_filter(model, iter, is_failed);
  case 3: //Show only runlevels
    gtk_tree_model_get(model, iter, 1, &str_data, -1);
    if(str_data && !strcmp(str_data, "runlevel")) {
      g_free(str_data);
      return TRUE;
    }
    else if(str_data){
      g_free(str_data);
    }
    return FALSE;
  default: //Show all (0) or nothing choosen (-1)
    return TRUE;
  }
}

void servicelist_cleanup() {
  GtkTreeView *service_list = GTK_TREE_VIEW(glade_xml_get_widget(xml, "service_list"));
  GtkTreeModelSort *sortmodel = GTK_TREE_MODEL_SORT(gtk_tree_view_get_model(service_list));
  GtkTreeModelFilter *filtermodel = GTK_TREE_MODEL_FILTER(gtk_tree_model_sort_get_model(sortmodel));
  GtkTreeModel *model = gtk_tree_model_filter_get_model(filtermodel);
  GtkTreeIter iter;
  struct s_info *serviceinfo;

  gboolean valid = gtk_tree_model_get_iter_first(model, &iter);
  while(valid) {
    gtk_tree_model_get(model, &iter, 4, &serviceinfo, -1);
    if(serviceinfo) {
      g_free(serviceinfo->startline);
      if(serviceinfo->fullname)
	g_free(serviceinfo->fullname);
      if(serviceinfo->file_content) {
        serviceinfo->file_content->references--;
        if(serviceinfo->file_content->references == 0) {
       	  if(serviceinfo->file_content->data)
	    g_free(serviceinfo->file_content->data);
	  if(serviceinfo->file_content->path)
	    g_free(serviceinfo->file_content->path);
	  g_free(serviceinfo->file_content);
        }
      }
      g_free(serviceinfo);
    }
    valid = gtk_tree_model_iter_next(model, &iter);
  }
}

char *path_tokenizer(char **name) {
  char *token = *name;
  char *rest = strchr(*name, '/');
  if(rest != NULL) {
    rest[0] = 0;
    rest++;
  }
  *name = rest;
  return token;
}

gboolean find_child(GtkTreeStore *store, GtkTreeIter *parent_iter,
                    GtkTreeIter *match_iter, char *name) {
  gchar *str_data;
  gboolean valid = gtk_tree_model_iter_children(GTK_TREE_MODEL(store),
                                                match_iter, parent_iter);

  while(valid) {
    gtk_tree_model_get(GTK_TREE_MODEL(store), match_iter, 0, &str_data, -1);
    if(str_data && !strcmp(str_data, name)) {
      g_free(str_data);
      return TRUE;
    }
    g_free(str_data);
    valid = gtk_tree_model_iter_next(GTK_TREE_MODEL(store), match_iter);
  }
  return FALSE;
}

gboolean add_recursive(GtkTreeStore *store, GtkTreeIter *parent_iter,
                   GtkTreeIter *iter, const char *name) {
  char *rest = strdup(name);
  char *tok = path_tokenizer(&rest);
  gboolean was_added = FALSE;

  if(!gtk_tree_model_iter_children(GTK_TREE_MODEL(store), iter, parent_iter) ||
     !find_child(store, parent_iter, iter, tok)) { //The node has to be created
    was_added = TRUE;
    gtk_tree_store_prepend(store, iter, parent_iter);
    gtk_tree_store_set(store, iter, 0, tok, 2, NULL, 4, NULL, -1);
  }

  if(rest) {
    parent_iter = gtk_tree_iter_copy(iter);
    was_added = add_recursive(store, parent_iter, iter, rest);
    gtk_tree_iter_free(parent_iter);
  }
  free(tok);
  return was_added;
}

gboolean add_node(GtkTreeStore *store, GtkTreeIter *iter, const char *name) {
  return add_recursive(store, NULL, iter, name);
}

void recursive_scan(char* path, GtkTreeStore *store) {
  GtkTreeIter iter;
  struct s_info *serviceinfo;
  struct s_file_content *file_content;
  
  struct dirent **namelist;
  int n = scandir(path, &namelist, &filefilter, alphasort);
  if(n < 0)
    perror(_("Unable to read directory"));
  else {
    FILE *file, *stream;
    while(n--) {
      struct stat buff;
      char *fullpath = malloc(strlen(path) + strlen(namelist[n]->d_name) + 2);
      sprintf(fullpath, "%s/%s", path, namelist[n]->d_name);
      if(lstat(fullpath, &buff)<0)
        break;
      if(S_ISDIR(buff.st_mode)) {
        recursive_scan(fullpath, store);
	free(fullpath);
      } else if(!fnmatch("*.i", namelist[n]->d_name, 0)) {
	file_content = malloc(sizeof(struct s_file_content));
	file_content->data = malloc(buff.st_size + 1);
        file = fopen(fullpath, "rt");   /* Open file. */
        if (file == NULL)
          continue;
	fread(file_content->data, buff.st_size, 1, file);
	file_content->data[buff.st_size] = 0;
	file_content->needs_saving = FALSE;
	file_content->path = fullpath;
	file_content->references = 0;
	stream = fmemopen(file_content->data, strlen(file_content->data), "r");
        char* buffer = malloc(255); //UGLY!
        while(fgets(buffer, 255, stream) > 0) {
          if(strstr(buffer, "daemon") == buffer) {
            serviceinfo = malloc(sizeof(struct s_info));
            serviceinfo->startline = malloc(strlen(buffer) + 1);
            strcpy(serviceinfo->startline, buffer);
            char *tok = strtok(buffer, " ");
            tok = strtok(NULL, " ");
	    add_node(store, &iter, tok);
            gtk_tree_store_set (store, &iter, 1, "daemon", 2, NULL, 4, serviceinfo, -1);
            serviceinfo->fullname = malloc(strlen(tok) + 1);
            strcpy(serviceinfo->fullname, tok);
            serviceinfo->type = DAEMON;
	    serviceinfo->file_content = file_content;
	    file_content->references++;
          } else if(strstr(buffer, "service") == buffer) {
            serviceinfo = malloc(sizeof(struct s_info));
            serviceinfo->startline = malloc(strlen(buffer) + 1);
            strcpy(serviceinfo->startline, buffer);
            char *tok = strtok(buffer, " ");
            tok = strtok(NULL, " ");
            add_node(store, &iter, tok);
            gtk_tree_store_set (store, &iter, 1, "service", 2, NULL, 4, serviceinfo, -1);
            serviceinfo->fullname = malloc(strlen(tok) + 1);
            strcpy(serviceinfo->fullname, tok);;
            serviceinfo->type = SERVICE;
	    serviceinfo->file_content = file_content;
	    file_content->references++;
          } else if(strstr(buffer, "virtual") == buffer) {
            serviceinfo = malloc(sizeof(struct s_info));
            serviceinfo->startline = malloc(strlen(buffer) + 1);
            strcpy(serviceinfo->startline, buffer);
            char *tok = strtok(buffer, " ");
            tok = strtok(NULL, " ");
            add_node(store, &iter, tok);
            gtk_tree_store_set (store, &iter, 1, "virtual", 2, NULL, 4, serviceinfo, -1);
            serviceinfo->fullname = malloc(strlen(tok) + 1);
            strcpy(serviceinfo->fullname, tok);
            serviceinfo->type = VIRTUAL;
	    serviceinfo->file_content = file_content;
	    file_content->references++;
          }
        }
        fclose(file);
        free(buffer);
      } else if(!fnmatch("*.runlevel", namelist[n]->d_name, 0)) {
	file_content = malloc(sizeof(struct s_file_content));
	file_content->data = malloc(buff.st_size + 1);
        file = fopen(fullpath, "rt");
        if (file == NULL)
          continue;
	fread(file_content->data, buff.st_size, 1, file);
	file_content->data[buff.st_size] = 0;
	file_content->needs_saving = FALSE;
	file_content->path = fullpath;
	file_content->references = 1;
        serviceinfo = malloc(sizeof(struct s_info));
        namelist[n]->d_name[strlen(namelist[n]->d_name)-9] = 0;
        add_node(store, &iter, namelist[n]->d_name);
        gtk_tree_store_set(store, &iter, 1, "runlevel", 2, NULL, 4, serviceinfo, -1);
        serviceinfo->fullname = malloc(strlen(namelist[n]->d_name) + 1);
        strcpy(serviceinfo->fullname, namelist[n]->d_name);
        serviceinfo->type = RUNLEVEL;
        serviceinfo->startline = NULL;
	serviceinfo->file_content = file_content;
      } else if(!fnmatch("*.virtual", namelist[n]->d_name, 0)) {
	file_content = malloc(sizeof(struct s_file_content));
	file_content->data = malloc(buff.st_size + 1);
        file = fopen(fullpath, "rt");
        if (file == NULL)
          continue;
	fread(file_content->data, buff.st_size, 1, file);
	file_content->data[buff.st_size] = 0;
	file_content->needs_saving = FALSE;
	file_content->path = fullpath;
	file_content->references = 1;
        serviceinfo = malloc(sizeof(struct s_info));
        namelist[n]->d_name[strlen(namelist[n]->d_name)-8] = 0;
        add_node(store, &iter, namelist[n]->d_name);
        gtk_tree_store_set(store, &iter, 1, "virtual", 2, NULL, 4, serviceinfo, -1);
	serviceinfo->fullname = malloc(strlen(namelist[n]->d_name) + 1);
        strcpy(serviceinfo->fullname, namelist[n]->d_name);
        serviceinfo->type = VIRTUALFILE;
        serviceinfo->startline = NULL;
	serviceinfo->file_content = file_content;
      } else {
	free(fullpath);
      }
      free(namelist[n]);
    }
    free(namelist);
  }
}

gboolean show_service(GtkWidget *treeview) {
  GtkWidget *service_textview = glade_xml_get_widget(xml_show_service, "service_textview");
  GtkTreeSelection *selected = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));
  GtkTextBuffer *buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(service_textview));
  GtkWidget *show_service = glade_xml_get_widget(xml_show_service, "show_service");

  GtkTreeIter iter;
  GtkTreeModel *model;
  FILE *stream;
  gchar res;
  int depth = 1;
  GtkTextIter start, end;
  
  gtk_tree_selection_get_selected(selected, &model, &iter);
  struct s_info *serviceinfo;
  gtk_tree_model_get(model, &iter, 4, &serviceinfo, -1);
  if(!serviceinfo) {
    gtk_text_buffer_set_text(buffer, _("Unable to locate service!"), -1);
    return FALSE;
  }
  struct s_file_content *file_content = serviceinfo->file_content;

  if(!file_content) {
    gtk_text_buffer_set_text(buffer, _("Don't know where to locate file!"), -1);
    return FALSE;
  }
  stream = fmemopen(file_content->data, strlen(file_content->data), "r");
  if (stream == NULL) {
    gtk_text_buffer_set_text(buffer, _("Couldn't open file!"), -1);
    return FALSE;
  }
  
  if(serviceinfo->type != RUNLEVEL && serviceinfo->type != VIRTUALFILE) {
    char* temp = malloc(255); //UGLY!
    while(fgets(temp, 255, stream) > 0) {
      if(strcmp(temp, serviceinfo->startline) == 0) {
        break;
      }
    }
    free(temp);
  }

  gtk_text_buffer_set_text(buffer, "", 0);
  
  do {
    res = fgetc(stream);
    if(res == '{') {
      depth++;
    } else if(res == '}') {
      depth--;
    }
    if(depth <= 0 || res == EOF)
      break;
    gtk_text_buffer_insert_at_cursor(buffer, &res, 1);
  } while(TRUE);
  gtk_text_buffer_get_bounds(buffer, &start, &end);

  gtk_window_set_title(GTK_WINDOW(show_service), serviceinfo->fullname);

  return TRUE;
}

void change_service() {
  GtkWidget *service_textview = glade_xml_get_widget(xml_show_service, "service_textview");
  GtkWidget *service_list = glade_xml_get_widget(xml, "service_list");
  GtkTreeSelection *selected = gtk_tree_view_get_selection(GTK_TREE_VIEW(service_list));
  GtkTextBuffer *buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(service_textview));
  GtkTreeIter iter;
  GtkTreeModel *model;
  gchar *servicename;
  gchar *newdata;
  FILE *oldstream, *newstream;
  size_t newsize;
  gchar res;
  int depth = 0;
  GtkTextIter start, end;
  
  gtk_tree_selection_get_selected(selected, &model, &iter);
  struct s_info *serviceinfo;
  gtk_tree_model_get(model, &iter, 0, &servicename, 4, &serviceinfo, -1);
  struct s_file_content *file_content = serviceinfo->file_content;

  oldstream = fmemopen(file_content->data, strlen(file_content->data), "r");
  if (oldstream == NULL)
    return;
  newstream = open_memstream(&newdata, &newsize);
  
  //Put anything before the current service into the file
  if(serviceinfo->type != RUNLEVEL && serviceinfo->type != VIRTUALFILE) {
    char* temp = malloc(255); //UGLY!
    while(fgets(temp, 255, oldstream) > 0) {
      fputs(temp, newstream);
      if(strcmp(temp, serviceinfo->startline) == 0) {
        break;
      }
    }
    free(temp);
    if(strrchr(serviceinfo->startline, '{'))
      depth++;
  }

  //Put the old contents of the service into the trash
  do {
    res = fgetc(oldstream);
    if(res == '{') {
      depth++;
    } else if(res == '}') {
      depth--;
    }
    if(depth <= 0 || res == EOF)
      break;
  } while(TRUE);

  //Put the new contents of the service into the file
  gtk_text_buffer_get_bounds(buffer, &start, &end);
  gchar *temp = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);
  fputs(temp, newstream);
  g_free(temp);

  //Put anything after the current service into the file
  if(serviceinfo->type != RUNLEVEL && serviceinfo->type != VIRTUALFILE) {
    fputs("}", newstream);
    do {
      res = fgetc(oldstream);
      if(res == EOF)
	break;
      fputc(res, newstream);
    } while(TRUE);
  }

  fclose(oldstream);
  fclose(newstream);

  free(file_content->data);

  file_content->data = newdata;
  file_content->needs_saving = TRUE;

  g_free(servicename);
}

gboolean save_service(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data) {
  FILE *file;
  struct s_info *serviceinfo;
  gtk_tree_model_get(model, iter, 4, &serviceinfo, -1);
  if(!serviceinfo)
    return FALSE;
  struct s_file_content *file_content = serviceinfo->file_content;
  if(file_content && file_content->needs_saving) {
    file = fopen(file_content->path, "w");
    if (file != NULL) {
      fputs(file_content->data, file);
      file_content->needs_saving = FALSE;
      fclose(file);
    }
  }
  return FALSE;
}

void save_all_services() {
  GtkTreeView *service_list = GTK_TREE_VIEW(glade_xml_get_widget(xml, "service_list"));
  GtkTreeModel *model = gtk_tree_view_get_model(service_list);

  gtk_tree_model_foreach(model, save_service, NULL);
}

gint service_sorter(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer user_data) {
  gint sortcol = GPOINTER_TO_INT(user_data);
  gint ret = 0;
  gchar *name1, *name2;

  gtk_tree_model_get(model, a, sortcol, &name1, -1);
  gtk_tree_model_get(model, b, sortcol, &name2, -1);

  ret = g_utf8_collate(name1,name2);

  g_free(name1);
  g_free(name2);

  return ret;
}

void populate_service_list() {
  gdk_threads_enter();
  GtkWidget *service_list = glade_xml_get_widget(xml, "service_list");
  GtkTreeModel *filtermodel, *sortmodel;
  GtkTreeViewColumn *column;
  GtkTreeStore  *servicestore = gtk_tree_store_new (5, G_TYPE_STRING, G_TYPE_STRING,
						    G_TYPE_STRING, G_TYPE_STRING,
						    G_TYPE_POINTER);

  //Column containing names
  column = gtk_tree_view_column_new_with_attributes(_("Name"),
						    GTK_CELL_RENDERER(gtk_cell_renderer_text_new()),
						    "text", 0, "background", 3, NULL);
  gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(servicestore), 0, service_sorter, 
				  GINT_TO_POINTER(0), NULL);
  gtk_tree_view_column_set_sort_column_id(column, 0);
  gtk_tree_view_insert_column(GTK_TREE_VIEW(service_list), column, -1);
  //Column containing function (runlevel/service/daemon/virtual)
  column = gtk_tree_view_column_new_with_attributes(_("Function"),
						    GTK_CELL_RENDERER(gtk_cell_renderer_text_new()),
						    "text", 1, "background", 3, NULL);
  gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(servicestore), 1, service_sorter, 
				  GINT_TO_POINTER(1), NULL);
  gtk_tree_view_column_set_sort_column_id(column, 1);
  gtk_tree_view_insert_column(GTK_TREE_VIEW(service_list), column, -1);
  //Column containing state
  column = gtk_tree_view_column_new_with_attributes(_("State"),
						    GTK_CELL_RENDERER(gtk_cell_renderer_text_new()),
						    "text", 2, "background", 3, NULL);
  gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(servicestore), 2, service_sorter, 
				  GINT_TO_POINTER(2), NULL);
  gtk_tree_view_column_set_sort_column_id(column, 2);
  gtk_tree_view_insert_column(GTK_TREE_VIEW(service_list), column, -1);
  //Hidden column containing color information
  column = gtk_tree_view_column_new_with_attributes("color",
						    GTK_CELL_RENDERER(gtk_cell_renderer_text_new()),
						    "text", 3,NULL);
  gtk_tree_view_column_set_visible(column, FALSE);
  gtk_tree_view_insert_column(GTK_TREE_VIEW(service_list), column, -1);

  recursive_scan(BASEDIR, servicestore);
  //service_list -> sortmodel -> filtermodel -> servicestore
  filtermodel = gtk_tree_model_filter_new(GTK_TREE_MODEL(servicestore), NULL);
  gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER(filtermodel),
					 service_filter, NULL, NULL);
  sortmodel = gtk_tree_model_sort_new_with_model(filtermodel);
  gtk_tree_view_set_model(GTK_TREE_VIEW(service_list), sortmodel);
  gdk_threads_leave();
}
